﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace KingAndPrisoner
{
    class Program
    {

        static void Main(string[] args)
        {
            var prison = new Prison(100);
            prison.Run();
            int len = prison.Prisoners.Length;
            for (int x = 0; x < len; x++)
            {
                if (!prison.Prisoners[x].LightOn)
                {
                    Console.WriteLine("Wrong!!!囚犯({0})还没有出来放风过", x);
                    Console.ReadLine();
                }
            }
        }
    }

    class Helper
    {
        static bool DisabledWait = true;

        public static void Print(string msg, params object[] args)
        {
            Print(false, msg, args);
        }

        public static void Print(bool wait, string msg, params object[] args)
        {
            Console.WriteLine(msg, args);
            if (wait && !DisabledWait)
            {
                Console.ReadLine();
            }
        }

    }

    /// <summary>
    /// 监狱
    /// </summary>
    public class Prison
    {
        /// <summary>
        /// 灯
        /// </summary>
        public Light Light { get; set; }

        /// <summary>
        /// 牢房
        /// </summary>
        public Prisoner[] Prisoners { get; set; }

        /// <summary>
        /// 随机数生成器
        /// </summary>
        Random random;

        /// <summary>
        /// 囚犯数量
        /// </summary>
        private int PrisonerCount { get; set; }

        /// <summary>
        /// 构造函数
        /// </summary>
        /// <param name="count">囚犯数量</param>
        public Prison(int count)
        {
            PrisonerCount = count;
            random = new Random(GetRandomSeed());

            //初始化灯,并将状态设置为关
            Light = new Light() { Status = LightStatus.OFF };

            //初始化牢房数组
            Prisoners = new Prisoner[PrisonerCount];
            for (int x = 0; x < PrisonerCount; x++)
            {
                Prisoners[x] = new Prisoner() { NO = x };
            }
        }

        public int Run()
        {
            SetCountPrisoner();
            bool finished = false;
            int count = 0;
            while (!finished)
            {
                count++;
                Helper.Print("{1}\t第 {0} 天", count, DateTime.Today.AddDays(count).ToString("yyyy-MM-dd"));
                int x = GetRandomNumber();
                var prisoner = Prisoners[x];
                prisoner.Release();
                prisoner.TrunLightOn(this.Light);
                if (prisoner is CountPrisoner)
                {
                    //ShowMsg("囚犯({0})是计数人", x);
                    var counter = prisoner as CountPrisoner;
                    counter.TrunLightOff(this.Light);
                    finished = counter.Finished;
                }
            }
            return count;
        }

        /// <summary>
        /// 随机设置一个囚犯作为计数人
        /// </summary>
        void SetCountPrisoner()
        {
            int num = GetRandomNumber();
            var prisoner = Prisoners[num];
            Prisoners[num] = new CountPrisoner(prisoner) { Total = PrisonerCount };
            Helper.Print(true, "计数人是:囚犯({0})", num);
        }

        /// <summary>
        /// 取得随机数
        /// </summary>
        /// <returns></returns>
        int GetRandomNumber()
        {
            return random.Next(0, PrisonerCount);
        }

        public int GetRandomSeed()
        {
            byte[] bytes = new byte[4];
            System.Security.Cryptography.RNGCryptoServiceProvider rng = new System.Security.Cryptography.RNGCryptoServiceProvider();
            rng.GetBytes(bytes);
            return BitConverter.ToInt32(bytes, 0);
        }
    }

    /// <summary>
    /// 灯
    /// </summary>
    public class Light
    {
        /// <summary>
        /// 灯的状态
        /// </summary>
        public LightStatus Status { get; set; }

        public void TurnOn()
        {
            if (this.Status == LightStatus.OFF)
            {
                this.Status = LightStatus.ON;
                Helper.Print("灯被打开了");
            }
        }

        public void TurnOff()
        {
            if (this.Status == LightStatus.ON)
            {
                this.Status = LightStatus.OFF;
                Helper.Print("灯被关闭了");
            }
        }
    }

    /// <summary>
    /// 灯的状态枚举
    /// </summary>
    public enum LightStatus { ON, OFF }

    /// <summary>
    /// 囚犯
    /// </summary>
    public class Prisoner
    {
        public int NO { get; set; }

        /// <summary>
        /// 是否开过灯了
        /// </summary>
        public bool LightOn { get; set; }

        public void Release()
        {
            Helper.Print("囚犯({0})出来放风", NO);
        }

        /// <summary>
        /// 将监狱的灯打开
        /// </summary>
        /// <param name="light"></param>
        public void TrunLightOn(Light light)
        {
            if (light.Status == LightStatus.ON)
            {
                Helper.Print("灯是开着的", this.NO);
                return;
            }
            if (LightOn)
            {
                Helper.Print("囚犯({0})已经开过灯了", NO);
                return;
            }
            LightOn = true;
            Helper.Print("囚犯({0})开灯", NO);
            light.TurnOn();
        }
    }

    /// <summary>
    /// 负责计数的囚犯
    /// </summary>
    public class CountPrisoner : Prisoner
    {
        /// <summary>
        /// 构造函数
        /// </summary>
        /// <param name="prisoner"></param>
        public CountPrisoner(Prisoner prisoner)
        {
            this.NO = prisoner.NO;
            this.LightOn = prisoner.LightOn;
        }

        /// <summary>
        /// 关灯计数
        /// </summary>
        public int Count { get; set; }

        public int Total { get; set; }

        public bool Finished { get; set; }

        /// <summary>
        /// 将监狱的灯关闭
        /// </summary>
        /// <param name="light"></param>
        public void TrunLightOff(Light light)
        {
            if (light.Status == LightStatus.ON)
            {
                Count++;
                Helper.Print("囚犯({0})关灯", this.NO);
                light.TurnOff();
                if (Count < Total)
                {
                    Helper.Print(true, "囚犯({0})已经统计到{1}个囚犯出来放风过了", this.NO, this.Count);
                }
                else
                {
                    Finished = true;
                    Helper.Print(true, "囚犯({0})已经统计到所有囚犯全部出来放风过了,任务完成了", this.NO);
                    Console.ReadLine();
                }
            }
        }
    }
}
